// DO NOT DELETE NEXT LINE
// LMA1.0 

/****************************************************************************

  TextEdit 2.x

  2.0  :          Source for TextEdit 2.0 from Tony Chang; no changelog provided
  2.1l :          Added in an escape code for the ; character.
  2.2l :          Changed from using " as string encapsulation to @
                  Added in an escape code for the * character.
                  Added in an escape code for the " character.
                  Added in an escape code for the @ character.
  2.2l Build 5 :  Added escape code for \\.
                  Added escape code for current build number.
  2.2l Build 6 :  Explorer version tab support added.
  2.3l Build 7 :  & escape code for * has been changed to > (preventing problems with \& constructs).
                  Compiled as multithreaded DLL.
                  Compiled against current LiteStep 0.24.6 CVS
  2.4l Build 11 : Re-written code to provide greater flexibility.
                  Added version status report for !About dialog.
                  Uses new version syntax now via TextEditNewSyntax
                  Supplemented TextDebug with TextEditDebug

****************************************************************************/

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <fstream.h>
#include "textedit.h"
#include <list>
#include <string>
using namespace std;

const char szAppName[] = "TextEdit"; // Our window class, etc
const char rcsRevision[] = "$Revision: 2.5 $"; // Our Version
const char rcsId[] = "$Id: textedit.cpp,v 2.5 Exp $"; // The Full RCS ID.
TextEdit *textedit; // The module

bool DEBUG;
bool DEBUG1;
bool DEBUG2;
bool SYNTAX;

//=========================================================
// Utility
//=========================================================
void DisplayMessage(LPCSTR pszMessage, LPCSTR pszTitle, UINT uExtraFlags = 0)
{
    MessageBox(NULL, pszMessage, szAppName,
        MB_OK | MB_TOPMOST | MB_SETFOREGROUND | uExtraFlags);
}

void DisplayMessage(LPCSTR pszMessage, UINT uExtraFlags = 0)
{
    DisplayMessage(pszMessage, szAppName, uExtraFlags);
}

//=========================================================
// Initialization and cleanup
//=========================================================
int initModuleEx(HWND ParentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
  int code;
  Window::init(dllInst); // missed by the module wizard (b3)
  textedit = new TextEdit(ParentWnd, code);
  return code;
}

void quitModule(HINSTANCE dllInst)
{
  delete textedit;
}


//=========================================================
// Bang commands
//=========================================================

void BangTextEdittextAppendFunction(HWND caller, const char* arg) {
  textedit->BangTextEdittextAppend(caller, arg);
}
void BangTextEdittextReplaceFunction(HWND caller, const char* arg) {
  textedit->BangTextEdittextReplace(caller, arg);
}
void BangTextEdittextDeleteFunction(HWND caller, const char* arg) {
  textedit->BangTextEdittextDelete(caller, arg);
}

//---------------------------------------------------------
// Module constructor
//---------------------------------------------------------
TextEdit::TextEdit(HWND parentWnd, int& code):
Window(szAppName)
{

  int msgs[] = {LM_GETREVID, 0};


  if (!createWindow(WS_EX_TOOLWINDOW, szAppName, WS_POPUP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
                    0, 0, 0, 0, parentWnd))
  {
    DisplayMessage("Unable to create window.", MB_ICONERROR);
    code = 1;
    return;
  }

  SetWindowLong(hWnd, GWL_USERDATA, magicDWord);

  SendMessage(hParent, LM_REGISTERMESSAGE, (WPARAM)hWnd, (LPARAM)msgs);

  DEBUG1 = GetRCBool("TextDebug", TRUE);
  DEBUG2 = GetRCBool("TextEditDebug", TRUE);
  if (DEBUG1 == TRUE || DEBUG2 == TRUE)
    DEBUG = TRUE;

  SYNTAX = GetRCBool("TextEditNewSyntax", TRUE);

  AddBangCommand("!textappend", BangTextEdittextAppendFunction);
  AddBangCommand("!textdel", BangTextEdittextDeleteFunction);
  AddBangCommand("!textreplace", BangTextEdittextReplaceFunction);

  code = 0;
}

//---------------------------------------------------------
// Module destructor
//---------------------------------------------------------

TextEdit::~TextEdit()
{
  int msgs[] = {LM_GETREVID, 0};

  SendMessage(hParent, LM_UNREGISTERMESSAGE, (WPARAM)hWnd, (LPARAM)msgs);

  RemoveBangCommand("!textappend");
  RemoveBangCommand("!textdel");
  RemoveBangCommand("!textreplace");

  destroyWindow();
}

//=========================================================
// Registered messages
//=========================================================

void TextEdit::windowProc(Message& message)
{
  BEGIN_MESSAGEPROC
    MESSAGE(onEndSession,        WM_ENDSESSION)
    MESSAGE(onEndSession,        WM_QUERYENDSESSION)
    MESSAGE(onGetRevId,          LM_GETREVID)
    MESSAGE(onSysCommand,        WM_SYSCOMMAND)
  END_MESSAGEPROC
}


//=========================================================
// Message handlers
//=========================================================
void TextEdit::onEndSession(Message& message)
{
  message.lResult = SendMessage(hParent, message.uMsg, message.wParam, message.lParam);
}

void TextEdit::onGetRevId(Message& message)
{
  char* buf = (char*)(message.lParam);

  switch (message.wParam)
  {
    case 0:
      sprintf(buf, "textedit.dll: %s", &rcsRevision[11]);
      buf[strlen(buf) - 1] = '\0';
      break;
    case 1:
      strcpy(buf, &rcsId[1]);
      buf[strlen(buf) - 1] = '\0';
      break;
    default:
      strcpy(buf, "");
  }
  message.lResult = strlen(buf);
}


void TextEdit::onSysCommand(Message& message)
{
  if (message.wParam == SC_CLOSE)
    PostMessage(hParent, WM_KEYDOWN, LM_SHUTDOWN, 0);
  else
    message.lResult = DefWindowProc(hWnd, message.uMsg, message.wParam, message.lParam);
}

//=========================================================
// getToken grabs a token and clobbers info that is parsed
//=========================================================
void getToken(char line[4096], char temp[4096])
{
	int offset = 0, i;
	bool endquote = false;
    char ENCAP;

    if (SYNTAX == TRUE)
	{
        ENCAP = '@';
	}
    else
	{
        ENCAP = '"';
	}

	while (line[offset] == ' ')
	{
		offset++;
	}

	if (line[offset] == ENCAP)
	{
		offset++;
		endquote = true;
	}

	for (i = 0; i+offset < int(strlen(line)); ++i)
	{
		// Copy else if block to add more escape codes.
		if (line[i+offset] == '\\' && int(strlen(line)) != i+offset+1)
		{
			if (line[i+offset+1] == ENCAP || line[i+offset+1] == '\\')
			{
				offset++;
				temp[i] = line[i+offset];
				continue;
			}
			// Bang escape code below. Present in 2.0 from Tony Chang
			else if (line[i+offset+1] == '^')
			{
				offset++;
				temp[i] = '!';
				continue;
			}
			// EVar escape code below. Present in 2.0 from Tony Chang
			else if (line[i+offset+1] == '#')
			{
				offset++;
				temp[i] = '$';
				continue;
			}
			// Comment escape code below. Added in 2.1l by leaf r&d
			else if (line[i+offset+1] == '~')
			{
				offset++;
				temp[i] = ';';
				continue;
			}
			// Asterisk escape code below. Added in 2.2l by leaf r&d
			   // Changed to > in 2.3l by leaf r&d
			else if (line[i+offset+1] == '>')
			{
				offset++;
				temp[i] = '*';
				continue;
			}
			// Quotation mark escape code below. Added in 2.2l by leaf r&d
			else if (line[i+offset+1] == '=')
			{
				offset++;
                temp[i] = '"';
                continue;
            }
            // Double slash escape code below.
            else if (line[i+offset+1] == '\\')
            {
                offset++;
                temp[i] = '\\';
                continue;
            }
            // Clipboard escape code
            else if (line[i+offset+1] == 'c')
            {
                if (OpenClipboard(NULL))
                {
                    HANDLE hData = GetClipboardData(CF_TEXT);
                    
                    if (hData)
                    {
                        char* pszBuffer = (char*)GlobalLock(hData);
                        temp[i] = '\0';
                        offset += min(strlen(pszBuffer), 4095-strlen(temp));
                        strncat(temp, pszBuffer, 4095-strlen(temp));
                        i = strlen(temp)-1;
                        GlobalUnlock(hData);
                    }
                    CloseClipboard();
                }
                else
                {
                    DisplayMessage(NULL, "Error: Couldn't open clipboard", MB_ICONERROR);
                }
            }
            // Debug needs to be kept at end of escape code chunk.
			else if (DEBUG==TRUE)
			{
				if (line[i+offset+1] < '0' || line[i+offset+1] > '9')
					DisplayMessage("Unknown escape code", MB_ICONERROR);
			}
		}

 	    if (line[i+offset] == ENCAP && endquote)
	  	    break;
		else if (line[i+offset] == ' ' && !endquote)
			break;
		else
			temp[i] = line[i+offset];
	}
	temp[i] = '\0';

	if (i+offset+1 > int(strlen(line)))
	{
		line[0] = '\0';
	}
	else
	{
		strcpy(line, line+i+offset+1);
	}
}

//=========================================================
// Bang command handling
//=========================================================

void TextEdit::BangTextEdittextAppend(HWND caller, const char* arg)
{
	char filename[4096], str[4096];
	char rest[4096];
	strcpy(rest, arg);

	getToken(rest, filename);
	getToken(rest, str);

	if (DEBUG==TRUE)
	{
		DisplayMessage(filename, "Filename");
		DisplayMessage(str, "String");
		DisplayMessage(rest, "Extra Chars");
	}

	::ofstream file;
	file.open(filename, /* ios::nocreate | */ ios::app);

	if (!file.fail())
	{
		file << str << endl;
	}
	else if (DEBUG==TRUE)
	{
		DisplayMessage("Failed to open file", MB_ICONERROR);
	}
	file.close();
}

void TextEdit::BangTextEdittextDelete(HWND caller, const char* arg)
{
	char filename[4096], regex[4096], input[4096];
	char rest[4096];
	regexp *r;
	strcpy(rest, arg);

	getToken(rest, filename);
	getToken(rest, regex);

	if (DEBUG==TRUE)
	{
		DisplayMessage(filename, "Filename");
		DisplayMessage(regex, "Regular Expression");
		DisplayMessage(rest, "Extra Chars");
	}

	r = regcomp(regex);

	::ifstream file;
	file.open(filename, ios::nocreate);

	if (!file.fail())
	{
		list<string> l;
		string s;
		while (file.getline(input, 4096))
		{
			if (!regexec(r, input))
			{
				s = input;
				l.push_back(s);
			}
			else if (DEBUG==TRUE)
			{
				DisplayMessage(input, "MATCH");
			}
		}
		file.close();
		::ofstream outfile;
		outfile.open(filename);
		for (list<string>::const_iterator lci = l.begin(); lci!=l.end(); lci++)
		{
			outfile << lci->data() << endl;
		}
		l.clear();
	}
	else if (DEBUG==TRUE)
	{
		file.close();
		DisplayMessage("Failed to open file", MB_ICONERROR);
	}
}

void TextEdit::BangTextEdittextReplace(HWND caller, const char* arg)
{
	char filename[4096], regex[4096], input[4096], replace[4096];
	char rest[4096], buf[4096];
	regexp *r;
	strcpy(rest, arg);

	getToken(rest, filename);
	getToken(rest, regex);
	getToken(rest, replace);

	if (DEBUG==TRUE)
	{
		DisplayMessage(filename, "Filename");
		DisplayMessage(regex, "Regular Expression");
		DisplayMessage(replace, "Replace string");
		DisplayMessage(rest, "Extra Chars");
	}

	r = regcomp(regex);

	::ifstream file;
	file.open(filename, ios::nocreate);

	if (!file.fail())
	{
		list<string> l;
		string s;
		while (file.getline(input, 4096))
		{
			if (regexec(r, input))
			{
				regsub(r, replace, buf);
				s = buf;
				l.push_back(buf);
				if (DEBUG==TRUE)
				{
					DisplayMessage(input, "MATCH");
					DisplayMessage(buf, "Replaced string");
				}
			}
			else
			{
				s = input;
				l.push_back(input);
			}
		}
		file.close();
		::ofstream outfile;
		outfile.open(filename);
		for (list<string>::const_iterator lci = l.begin(); lci!=l.end(); lci++)
		{
			outfile << lci->data() << endl;
		}
		l.clear();
	}
	else if (DEBUG==TRUE)
	{
		file.close();
		DisplayMessage("Failed to open file", MB_ICONERROR);
	}
}
